//
// Copyright 2021 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package options

import (
	"testing"

	"github.com/google/go-cmp/cmp"
)

func TestFlagToEnv(t *testing.T) {
	testCases := []struct {
		flag     string
		expected string
	}{
		{
			flag:     "rekor-url",
			expected: "COSIGN_REKOR_URL",
		},
		{
			flag:     "certificate",
			expected: "COSIGN_CERTIFICATE",
		},
		{
			flag:     "k8s-keychain",
			expected: "COSIGN_K8S_KEYCHAIN",
		},
		{
			flag:     "output-file",
			expected: "COSIGN_OUTPUT_FILE",
		},
		{
			flag:     "sbom",
			expected: "COSIGN_SBOM",
		},
	}

	for _, tc := range testCases {
		t.Run(tc.flag, func(t *testing.T) {
			result := flagToEnvVar(tc.flag)
			if diff := cmp.Diff(result, tc.expected); diff != "" {
				t.Fatal(diff)
			}
		})
	}
}
