//
// Copyright 2022 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package errors

type CosignError struct {
	Message string
	Code    int
}

func Error(cosignError CosignError) error {
	return &CosignError{
		Message: cosignError.Message,
		Code:    cosignError.Code,
	}
}

// Assert that we implement error at build time.
var _ error = (*CosignError)(nil)

// Error implements error
func (ce *CosignError) Error() string {
	return ce.Message
}

func (ce *CosignError) ExitCode() int {
	return ce.Code
}
